# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines a Builder API for Meridian `InputData`.

The Builder API for `InputData` exposes piecewise data ingestion with its own
validation logic and an overall final validation logic before a valid
`InputData` is constructed.
"""

import abc
from collections.abc import Sequence
import datetime
import warnings
from meridian import constants
from meridian.data import input_data
from meridian.data import time_coordinates as tc
import natsort
import numpy as np
import xarray as xr


__all__ = [
    'InputDataBuilder',
]


class InputDataBuilder(abc.ABC):
  """Abstract base class for `InputData` builders."""

  def __init__(self, kpi_type: str):
    self._kpi_type = kpi_type

    # These working attributes are going to be set along the way as the builder
    # is provided piecemeal with the user's input data.
    # In the course of processing each DataFrame piece, dimension coordinates
    # will be discovered and set with, e.g., `self.time_coords = ...`.
    # The setter code will perform basic validation
    # checks, e.g.:
    # * If previous dataframe input already set it, then it should be consistent
    # * If not, set it for the first time.
    # * When setting, make consistency checks against other dimensions
    # * etc...

    # Working dimensions and their coordinates.
    self._time_coords: Sequence[str] = None
    self._media_time_coords: Sequence[str] = None
    self._geos: Sequence[str] = None

    # Working data arrays (components of the final `InputData` object)
    self._kpi: xr.DataArray = None
    self._controls: xr.DataArray = None
    self._population: xr.DataArray = None
    self._revenue_per_kpi: xr.DataArray = None
    self._media: xr.DataArray = None
    self._media_spend: xr.DataArray = None
    self._reach: xr.DataArray = None
    self._frequency: xr.DataArray = None
    self._rf_spend: xr.DataArray = None
    self._organic_media: xr.DataArray = None
    self._organic_reach: xr.DataArray = None
    self._organic_frequency: xr.DataArray = None
    self._non_media_treatments: xr.DataArray = None

  @property
  def time_coords(self) -> Sequence[str]:
    return self._time_coords

  @time_coords.setter
  def time_coords(self, value: Sequence[str]):
    if len(value) != len(set(value)):
      raise ValueError('`times` coords must be unique.')
    if self.time_coords is not None and set(self.time_coords) != set(value):
      raise ValueError(f'`times` coords already set to {self.time_coords}.')
    if self.media_time_coords is not None and not set(value).issubset(
        self.media_time_coords
    ):
      raise ValueError(
          '`times` coords must be subset of previously set `media_times`'
          ' coords.'
      )
    if self.media_time_coords is not None:
      self._validate_lagged_media(
          media_time_coords=self.media_time_coords, time_coords=value
      )
    _ = tc.TimeCoordinates.from_dates(sorted(value)).interval_days
    self._time_coords = value

  @property
  def media_time_coords(self) -> Sequence[str]:
    return self._media_time_coords

  @media_time_coords.setter
  def media_time_coords(self, value: Sequence[str]):
    if len(value) != len(set(value)):
      raise ValueError('`media_times` coords must be unique.')
    if self.media_time_coords is not None and set(
        self.media_time_coords
    ) != set(value):
      raise ValueError(
          f'`media_times` coords already set to {self.media_time_coords}.'
      )
    if self.time_coords is not None and not set(value).issuperset(
        self.time_coords
    ):
      raise ValueError(
          '`media_times` coords must be superset of previously set `times`'
          ' coords.'
      )
    if self.time_coords is not None:
      self._validate_lagged_media(
          media_time_coords=value, time_coords=self.time_coords
      )
    _ = tc.TimeCoordinates.from_dates(sorted(value)).interval_days
    self._media_time_coords = value

  @property
  def geos(self) -> Sequence[str]:
    return self._geos

  @geos.setter
  def geos(self, value: Sequence[str]):
    if len(value) != len(set(value)):
      raise ValueError('Geos must be unique.')
    if self.geos is not None and set(self.geos) != set(value):
      raise ValueError(
          f'geos already set to {self.geos}. Cannot reassign to {value}.'
      )
    self._geos = value

  @property
  def kpi(self) -> xr.DataArray:
    return self._kpi

  @kpi.setter
  def kpi(self, kpi: xr.DataArray):
    """Sets the `kpi` data array.

    `kpi` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='kpi',
        dims=['geo', 'time'],
        coords={
          'geo': ...,
          'time': ...,
        },
    )
    ```

    Args:
      kpi: Kpi DataArray.
    """
    self._validate_set('KPI', self.kpi)

    self._kpi = self._normalize_coords(kpi, constants.TIME)
    self.geos = self.kpi.coords[constants.GEO].values.tolist()
    self.time_coords = self.kpi.coords[constants.TIME].values.tolist()

  @property
  def controls(self) -> xr.DataArray:
    return self._controls

  @controls.setter
  def controls(self, controls: xr.DataArray):
    """Sets the `controls` data array.

    `controls` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='controls',
        dims=['geo', 'time', 'control_variable'],
        coords={
          'geo': ...,
          'time': ...,
          'control_variable': ...,
        },
    )
    ```

    Args:
      controls: Controls DataArray.
    """
    self._validate_set('Controls', self.controls)

    self._controls = self._normalize_coords(controls, constants.TIME)
    self.geos = self.controls.coords[constants.GEO].values.tolist()
    self.time_coords = self.controls.coords[constants.TIME].values.tolist()

  @property
  def population(self) -> xr.DataArray:
    return self._population

  @population.setter
  def population(self, population: xr.DataArray):
    """Sets the `media` data array.

    `population` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='population',
        dims=['geo'],
        coords={
          'geo': ...,
        },
    )
    ```

    Args:
      population: Population DataArray.
    """
    self._validate_set('Population', self.population)

    self._population = self._normalize_coords(population)
    self.geos = self.population.coords[constants.GEO].values.tolist()

  @property
  def revenue_per_kpi(self) -> xr.DataArray:
    return self._revenue_per_kpi

  @revenue_per_kpi.setter
  def revenue_per_kpi(self, revenue_per_kpi: xr.DataArray):
    """Sets the `revenue_per_kpi` data array.

    `revenue_per_kpi` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='revenue_per_kpi',
        dims=['geo', 'time'],
        coords={
          'geo': ...,
          'time': ...,
        },
    )
    ```
    Args:
      revenue_per_kpi: Revenue per kpi DataArray.
    """
    self._validate_set('Revenue per KPI', self.revenue_per_kpi)

    self._revenue_per_kpi = self._normalize_coords(
        revenue_per_kpi, constants.TIME
    )
    self.geos = self.revenue_per_kpi.coords[constants.GEO].values.tolist()
    self.time_coords = self.revenue_per_kpi.coords[
        constants.TIME
    ].values.tolist()

  @property
  def media(self) -> xr.DataArray:
    return self._media

  @media.setter
  def media(self, media: xr.DataArray):
    """Sets the `media` data array.

    `media` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='media',
        dims=['geo', 'media_time', 'media_channel'],
        coords={
          'geo': ...,
          'media_time': ...,
          'media_channel': ...,
        },
    )
    ```

    Args:
      media: Media DataArray.
    """
    self._validate_set('Media', self.media)
    self._validate_channels_consistency(
        constants.MEDIA_CHANNEL, [media, self.media_spend]
    )

    self._media = self._normalize_coords(media, constants.MEDIA_TIME)
    self.geos = self.media.coords[constants.GEO].values.tolist()
    self.media_time_coords = self.media.coords[
        constants.MEDIA_TIME
    ].values.tolist()

  @property
  def media_spend(self) -> xr.DataArray:
    return self._media_spend

  @media_spend.setter
  def media_spend(self, media_spend: xr.DataArray):
    """Sets the `media_spend` data array.

    `media_spend` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='media_spend',
        dims=['geo', 'time', 'media_channel'],
        coords={
          'geo': ...,
          'time': ...,
          'media_channel': ...,
        },
    )
    ```

    Args:
      media_spend: Media spend DataArray.
    """
    self._validate_set('Media spend', self.media_spend)
    self._validate_channels_consistency(
        constants.MEDIA_CHANNEL, [media_spend, self.media]
    )

    self._media_spend = self._normalize_coords(media_spend, constants.TIME)
    self.geos = self.media_spend.coords[constants.GEO].values.tolist()
    self.time_coords = self.media_spend.coords[constants.TIME].values.tolist()

  @property
  def reach(self) -> xr.DataArray:
    return self._reach

  @reach.setter
  def reach(self, reach: xr.DataArray):
    """Sets the `reach` data array.

    `reach` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='reach',
        dims=['geo', 'media_time', 'rf_channel'],
        coords={
          'geo': ...,
          'media_time': ...,
          'rf_channel': ...,
        },
    )
    ```

    Args:
      reach: Reach DataArray.
    """
    self._validate_set('Reach', self.reach)
    self._validate_channels_consistency(
        constants.RF_CHANNEL, [reach, self.frequency, self.rf_spend]
    )

    self._reach = self._normalize_coords(reach, constants.MEDIA_TIME)
    self.geos = self.reach.coords[constants.GEO].values.tolist()
    self.media_time_coords = self.reach.coords[
        constants.MEDIA_TIME
    ].values.tolist()

  @property
  def frequency(self) -> xr.DataArray:
    return self._frequency

  @frequency.setter
  def frequency(self, frequency: xr.DataArray):
    """Sets the `frequency` data array.

    `frequency` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='frequency',
        dims=['geo', 'media_time', 'rf_channel'],
        coords={
          'geo': ...,
          'media_time': ...,
          'rf_channel': ...,
        },
    )
    ```

    Args:
      frequency: Frequency DataArray.
    """
    self._validate_set('Frequency', self.frequency)
    self._validate_channels_consistency(
        constants.RF_CHANNEL, [frequency, self.reach, self.rf_spend]
    )

    self._frequency = self._normalize_coords(frequency, constants.MEDIA_TIME)
    self.geos = self.frequency.coords[constants.GEO].values.tolist()
    self.media_time_coords = self.frequency.coords[
        constants.MEDIA_TIME
    ].values.tolist()

  @property
  def rf_spend(self) -> xr.DataArray:
    return self._rf_spend

  @rf_spend.setter
  def rf_spend(self, rf_spend: xr.DataArray):
    """Sets the `rf_spend` data array.

    `rf_spend` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='rf_spend',
        dims=['geo', 'time', 'rf_channel'],
        coords={
          'geo': ...,
          'time': ...,
          'rf_channel': ...,
        },
    )
    ```

    Args:
      rf_spend: RF spend DataArray.
    """
    self._validate_set('RF spend', self.rf_spend)
    self._validate_channels_consistency(
        constants.RF_CHANNEL, [rf_spend, self.reach, self.frequency]
    )

    self._rf_spend = self._normalize_coords(rf_spend, constants.TIME)
    self.geos = self.rf_spend.coords[constants.GEO].values.tolist()
    self.time_coords = self.rf_spend.coords[constants.TIME].values.tolist()

  @property
  def organic_media(self) -> xr.DataArray:
    return self._organic_media

  @organic_media.setter
  def organic_media(self, organic_media: xr.DataArray):
    """Sets the `organic_media` data array.

    `organic_media` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='organic_media',
        dims=['geo', 'media_time'],
        coords={
          'geo': ...,
          'media_time': ...,
        },
    )
    ```

    Args:
      organic_media: Organic media DataArray.
    """
    self._validate_set('Organic media', self.organic_media)

    self._organic_media = self._normalize_coords(
        organic_media, constants.MEDIA_TIME
    )
    self.geos = self.organic_media.coords[constants.GEO].values.tolist()
    self.media_time_coords = self.organic_media.coords[
        constants.MEDIA_TIME
    ].values.tolist()

  @property
  def organic_reach(self) -> xr.DataArray:
    return self._organic_reach

  @organic_reach.setter
  def organic_reach(self, organic_reach: xr.DataArray):
    """Sets the `organic_reach` data array.

    `organic_reach` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='organic_reach',
        dims=['geo', 'media_time', 'organic_rf_channel'],
        coords={
          'geo': ...,
          'media_time': ...,
          'organic_rf_channel': ...,
        },
    )
    ```

    Args:
      organic_reach: Organic reach DataArray.
    """
    self._validate_set('Organic reach', self.organic_reach)
    self._validate_channels_consistency(
        constants.ORGANIC_RF_CHANNEL, [organic_reach, self.organic_frequency]
    )

    self._organic_reach = self._normalize_coords(
        organic_reach, constants.MEDIA_TIME
    )
    self.geos = self.organic_reach.coords[constants.GEO].values.tolist()
    self.media_time_coords = self.organic_reach.coords[
        constants.MEDIA_TIME
    ].values.tolist()

  @property
  def organic_frequency(self) -> xr.DataArray:
    return self._organic_frequency

  @organic_frequency.setter
  def organic_frequency(self, organic_frequency: xr.DataArray):
    """Sets the `organic_frequency` data array.

    `organic_frequency` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='organic_frequency',
        dims=['geo', 'media_time', 'organic_rf_channel'],
        coords={
          'geo': ...,
          'media_time': ...,
          'organic_rf_channel': ...,
        },
    )
    ```

    Args:
      organic_frequency: Organic frequency DataArray.
    """
    self._validate_set('Organic frequency', self.organic_frequency)
    self._validate_channels_consistency(
        constants.ORGANIC_RF_CHANNEL, [organic_frequency, self.organic_reach]
    )

    self._organic_frequency = self._normalize_coords(
        organic_frequency, constants.MEDIA_TIME
    )
    self.geos = self.organic_frequency.coords[constants.GEO].values.tolist()
    self.media_time_coords = self.organic_frequency.coords[
        constants.MEDIA_TIME
    ].values.tolist()

  @property
  def non_media_treatments(self) -> xr.DataArray:
    return self._non_media_treatments

  @non_media_treatments.setter
  def non_media_treatments(self, non_media_treatments: xr.DataArray):
    """Sets the `non media treatments` data array.

    `non_media_treatments` must have the following `DataArray` signature:

    ```
    xarray.DataArray(
        data=...,
        name='non_media_treatments',
        dims=['geo', 'time', 'non_media_channel'],
        coords={
          'geo': ...,
          'time': ...,
          'non_media_channel': ...,
        },
    )
    ```

    Args:
      non_media_treatments: Non-media treatments DataArray.
    """
    self._validate_set('Non-media treatments', self.non_media_treatments)

    self._non_media_treatments = self._normalize_coords(
        non_media_treatments, constants.TIME
    )
    self.geos = self.non_media_treatments.coords[constants.GEO].values.tolist()
    self.time_coords = self.non_media_treatments.coords[
        constants.TIME
    ].values.tolist()

  def build(self) -> input_data.InputData:
    """Builds an `InputData`.

    Constructs an `InputData` from constituent `DataArray`s given to this
    builder thus far after performing one final validation pass over all data
    arrays for consistency checks.

    Returns:
      A validated `InputData`.
    """
    self._validate_required_components()
    self._validate_nas()

    # TODO: move logic from input_data to here: all channel names
    # should be unique across media channels, rf channels, organic media
    # channels, and organic rf channels.
    sorted_geos = natsort.natsorted(self.geos)
    sorted_times = natsort.natsorted(self.time_coords)
    sorted_media_times = natsort.natsorted(self.media_time_coords)

    def _get_sorted(da: xr.DataArray | None, is_media_time: bool = False):
      """Naturally sorts the DataArray by geo and time/media time."""

      if da is None:
        return None
      if is_media_time:
        return da.reindex(geo=sorted_geos, media_time=sorted_media_times)
      else:
        return da.reindex(geo=sorted_geos, time=sorted_times)

    return input_data.InputData(
        kpi_type=self._kpi_type,
        kpi=_get_sorted(self.kpi),
        revenue_per_kpi=_get_sorted(self.revenue_per_kpi),
        controls=_get_sorted(self.controls),
        population=self.population.reindex(geo=sorted_geos),
        media=_get_sorted(self.media, True),
        media_spend=_get_sorted(self.media_spend),
        reach=_get_sorted(self.reach, True),
        frequency=_get_sorted(self.frequency, True),
        rf_spend=_get_sorted(self.rf_spend),
        non_media_treatments=_get_sorted(self.non_media_treatments),
        organic_media=_get_sorted(self.organic_media, True),
        organic_reach=_get_sorted(self.organic_reach, True),
        organic_frequency=_get_sorted(self.organic_frequency, True),
    )

  def _normalize_coords(
      self, da: xr.DataArray, time_dimension_name: str | None = None
  ) -> xr.DataArray:
    """Normalizes the given `DataArray`'s coordinates in Meridian convention.

    Validates that time values are in the conventional Meridian format and
    that geos have national name if national. If geo coordinates are not string-
    typed, they will be converted to strings.

    Args:
      da: The DataArray to normalize.
      time_dimension_name: The name of the time dimension. If None, the will
        skip time normalization.

    Returns:
      The normalized DataArray.
    """
    if time_dimension_name is not None:
      # Time values are expected to be
      # (a) strings formatted in `"yyyy-mm-dd"`
      # or
      # (b) `datetime` values as numpy's `datetime64` types.
      # All other types are not currently supported.

      # If (b), `datetime` coord values will be normalized as formatted strings.

      if da.coords.dtypes[time_dimension_name] == np.dtype('datetime64[ns]'):
        date_strvalues = np.datetime_as_string(
            da.coords[time_dimension_name], unit='D'
        )
        da = da.assign_coords({time_dimension_name: date_strvalues})

      # Assume that the time coordinate labels are date-formatted strings.
      # We don't currently support other, arbitrary object types in the builder.
      for time in da.coords[time_dimension_name].values:
        try:
          _ = datetime.datetime.strptime(time, constants.DATE_FORMAT)
        except ValueError as exc:
          raise ValueError(
              f"Invalid time label: '{time}'. Expected format:"
              f" '{constants.DATE_FORMAT}'"
          ) from exc

    if len(da.coords[constants.GEO].values.tolist()) == 1:
      da = da.assign_coords(
          {constants.GEO: [constants.NATIONAL_MODEL_DEFAULT_GEO_NAME]},
      )
    else:
      da = da.assign_coords(
          {constants.GEO: da.coords[constants.GEO].astype(str)}
      )

    return da

  def _validate_set(self, component: str, da: xr.DataArray):
    if da is not None:
      raise ValueError(f'{component} was already set to {da}.')

  def _validate_channels_consistency(
      self, channel_dimension_name: str, da_list: list[xr.DataArray | None]
  ):
    for da in da_list:
      if da is not None and set(
          da.coords[channel_dimension_name].values.tolist()
      ) != set(da_list[0].coords[channel_dimension_name].values.tolist()):
        raise ValueError(
            f'{channel_dimension_name} coordinates must be the same between'
            f' {[da.name for da in da_list if da is not None]}.'
        )

  def _validate_required_components(self):
    """Validates that all required data arrays are defined."""
    if self.kpi is None:
      raise ValueError('KPI is required.')

    if len(self.geos) == 1:
      if self.population is not None:
        warnings.warn(
            'The `population` argument is ignored in a nationally aggregated'
            ' model. It will be reset to [1, 1, ..., 1]'
        )
      self._population = xr.DataArray(
          [constants.NATIONAL_MODEL_DEFAULT_POPULATION_VALUE],
          dims=[constants.GEO],
          coords={
              constants.GEO: self.geos,
          },
          name=constants.POPULATION,
      )
    if self.population is None:
      raise ValueError('Population is required for non national models.')

    if (self.media is None) ^ (self.media_spend is None):
      raise ValueError('Media and media spend must be provided together.')
    if (
        self.reach is not None
        or self.frequency is not None
        or self.rf_spend is not None
    ) and (
        self.reach is None or self.frequency is None or self.rf_spend is None
    ):
      raise ValueError(
          'Reach, frequency, and rf_spend must be provided together.'
      )
    if (self.organic_reach is None) ^ (self.organic_frequency is None):
      raise ValueError(
          'Organic reach and organic frequency must be provided together.'
      )
    if (
        self.reach is None
        and self.frequency is None
        and self.rf_spend is None
        and self.media_spend is None
        and self.media is None
    ):
      raise ValueError(
          'It is required to have at least one of media or reach + frequency.'
      )

  def _validate_nas(self):
    """Check for NAs in all of the DataArrays.

    Since the DataArray components should already distinguish between media time
    and time coords, there are no media times to infer so there should be no
    NAs.
    """
    if self.kpi.isnull().any(axis=None):
      raise ValueError('NA values found in the kpi data.')
    if self.population.isnull().any(axis=None):
      raise ValueError('NA values found in the population data.')
    if self.controls is not None and self.controls.isnull().any(axis=None):
      raise ValueError('NA values found in the controls data.')
    if self.revenue_per_kpi is not None and self.revenue_per_kpi.isnull().any(
        axis=None
    ):
      raise ValueError('NA values found in the revenue per kpi data.')
    if self.media_spend is not None and self.media_spend.isnull().any(
        axis=None
    ):
      raise ValueError('NA values found in the media spend data.')
    if self.rf_spend is not None and self.rf_spend.isnull().any(axis=None):
      raise ValueError('NA values found in the rf spend data.')
    if (
        self.non_media_treatments is not None
        and self.non_media_treatments.isnull().any(axis=None)
    ):
      raise ValueError('NA values found in the non media treatments data.')

    if self.media is not None and self.media.isnull().any(axis=None):
      raise ValueError('NA values found in the media data.')

    if self.reach is not None and self.reach.isnull().any(axis=None):
      raise ValueError('NA values found in the reach data.')
    if self.frequency is not None and self.frequency.isnull().any(axis=None):
      raise ValueError('NA values found in the frequency data.')

    if self.organic_media is not None and self.organic_media.isnull().any(
        axis=None
    ):
      raise ValueError('NA values found in the organic media data.')

    if self.organic_reach is not None and self.organic_reach.isnull().any(
        axis=None
    ):
      raise ValueError('NA values found in the organic reach data.')
    if (
        self.organic_frequency is not None
        and self.organic_frequency.isnull().any(axis=None)
    ):
      raise ValueError('NA values found in the organic frequency data.')

  def _validate_lagged_media(
      self, media_time_coords: Sequence[str], time_coords: Sequence[str]
  ):
    na_period = np.sort(list(set(media_time_coords) - set(time_coords)))
    if not np.all(na_period == np.sort(media_time_coords)[: len(na_period)]):
      raise ValueError(
          "The 'lagged media' period (period with 100% NA values in all"
          f' non-media columns) {na_period} is not a continuous window'
          ' starting from the earliest time period.'
      )



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Serialization and deserialization of Meridian models."""

import abc
from typing import Generic, TypeVar


WireFormat = TypeVar("WireFormat")
PythonType = TypeVar("PythonType")


class Serde(Generic[WireFormat, PythonType], abc.ABC):
  """Serializes and deserializes a Python type into a wire format."""

  def serialize(self, obj: PythonType, **kwargs) -> WireFormat:
    """Serializes the given object into a wire format."""
    raise NotImplementedError()

  def deserialize(
      self, serialized: WireFormat, serialized_version: str = "", **kwargs
  ) -> PythonType:
    """Deserializes the given wire format into a Python object.

    Args:
      serialized: The serialized object.
      serialized_version: The version of the serialized object. This is used to
        handle changes in deserialization logic across different versions.
      **kwargs: Additional keyword arguments to pass to the deserialization
        function.

    Returns:
      The deserialized object.
    """
    raise NotImplementedError()



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Builds Looker Studio report URLs.

This package provides tools to construct URLs for Looker Studio reports that
embed data directly within the URL itself. This is achieved through the creation
of shareable, pre-configured reports without requiring a separate, pre-existing
data source.

The primary functionality is exposed through the `url_generator` module.

Typical Usage:

  1. Use `url_generator.create_report_url()` to create the complete URL, based
  on a `sheets.Spreadsheet` object.

Example:

```python
from lookerstudio.linkingapi import url_generator
from lookerstudio.converters import sheets

# Generate the URL
looker_studio_report_url = url_generator.create_report_url(
    url="some_url",
    id="some_id",
    sheet_id_by_tab_name={},
)
# The `looker_studio_report_url` can now be shared to open a pre-populated
# report.
```
"""

from scenarioplanner.linkingapi import constants
from scenarioplanner.linkingapi import url_generator



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core mathematical equations for the Meridian model.

This module defines the `ModelEquations` class, which encapsulates the stateless
mathematical functions used in the Meridian MMM. This includes the core model
definitions, such as adstock, hill, and other transformations used
during model fitting. It requires a `ModelContext` instance for data access.
"""

from collections.abc import Sequence
import numbers

from meridian import backend
from meridian import constants
from meridian.model import adstock_hill
from meridian.model import context


__all__ = [
    "ModelEquations",
]


class ModelEquations:
  """Provides core, stateless mathematical functions for Meridian MMM."""

  def __init__(self, model_context: context.ModelContext):
    self._context = model_context

  def adstock_hill_media(
      self,
      *,
      media: backend.Tensor,
      alpha: backend.Tensor,
      ec: backend.Tensor,
      slope: backend.Tensor,
      decay_functions: str | Sequence[str] = constants.GEOMETRIC_DECAY,
      n_times_output: int | None = None,
  ) -> backend.Tensor:
    """Transforms media or using Adstock and Hill functions in the desired order.

    Args:
      media: Tensor of dimensions `(n_geos, n_media_times, n_media_channels)`
        containing non-negative media execution values. Typically this is
        impressions, but it can be any metric, such as `media_spend`. Clicks are
        often used for paid search ads.
      alpha: Uniform distribution for Adstock and Hill calculations.
      ec: Shifted half-normal distribution for Adstock and Hill calculations.
      slope: Deterministic distribution for Adstock and Hill calculations.
      decay_functions: String or sequence of strings denoting the adstock decay
        function(s) for each channel. Default: 'geometric'.
      n_times_output: Number of time periods to output. This argument is
        optional when the number of time periods in `media` equals
        `n_media_times`, in which case `n_times_output` defaults to `n_times`.

    Returns:
      Tensor with dimensions `[..., n_geos, n_times, n_media_channels]`
      representing Adstock and Hill-transformed media.
    """
    if n_times_output is None and (
        media.shape[1] == self._context.n_media_times
    ):
      n_times_output = self._context.n_times
    elif n_times_output is None:
      raise ValueError(
          "n_times_output is required. This argument is only optional when "
          "`media` has a number of time periods equal to `n_media_times`."
      )

    adstock_transformer = adstock_hill.AdstockTransformer(
        alpha=alpha,
        max_lag=self._context.model_spec.max_lag,
        n_times_output=n_times_output,
        decay_functions=decay_functions,
    )
    hill_transformer = adstock_hill.HillTransformer(
        ec=ec,
        slope=slope,
    )
    transformers_list = (
        [hill_transformer, adstock_transformer]
        if self._context.model_spec.hill_before_adstock
        else [adstock_transformer, hill_transformer]
    )

    media_out = media
    for transformer in transformers_list:
      media_out = transformer.forward(media_out)
    return media_out

  def adstock_hill_rf(
      self,
      *,
      reach: backend.Tensor,
      frequency: backend.Tensor,
      alpha: backend.Tensor,
      ec: backend.Tensor,
      slope: backend.Tensor,
      decay_functions: str | Sequence[str] = constants.GEOMETRIC_DECAY,
      n_times_output: int | None = None,
  ) -> backend.Tensor:
    """Transforms reach and frequency (RF) using Hill and Adstock functions.

    Args:
      reach: Tensor of dimensions `(n_geos, n_media_times, n_rf_channels)`
        containing non-negative media for reach.
      frequency: Tensor of dimensions `(n_geos, n_media_times, n_rf_channels)`
        containing non-negative media for frequency.
      alpha: Uniform distribution for Adstock and Hill calculations.
      ec: Shifted half-normal distribution for Adstock and Hill calculations.
      slope: Deterministic distribution for Adstock and Hill calculations.
      decay_functions: String or sequence of strings denoting the adstock decay
        function(s) for each channel. Default: 'geometric'.
      n_times_output: Number of time periods to output. This argument is
        optional when the number of time periods in `reach` equals
        `n_media_times`, in which case `n_times_output` defaults to `n_times`.

    Returns:
      Tensor with dimensions `[..., n_geos, n_times, n_rf_channels]`
      representing Hill and Adstock-transformed RF.
    """
    if n_times_output is None and (
        reach.shape[1] == self._context.n_media_times
    ):
      n_times_output = self._context.n_times
    elif n_times_output is None:
      raise ValueError(
          "n_times_output is required. This argument is only optional when "
          "`reach` has a number of time periods equal to `n_media_times`."
      )

    hill_transformer = adstock_hill.HillTransformer(
        ec=ec,
        slope=slope,
    )
    adstock_transformer = adstock_hill.AdstockTransformer(
        alpha=alpha,
        max_lag=self._context.model_spec.max_lag,
        n_times_output=n_times_output,
        decay_functions=decay_functions,
    )
    adj_frequency = hill_transformer.forward(frequency)
    rf_out = adstock_transformer.forward(reach * adj_frequency)

    return rf_out

  def compute_non_media_treatments_baseline(
      self,
      non_media_baseline_values: Sequence[str | float] | None = None,
  ) -> backend.Tensor:
    """Computes the baseline for each non-media treatment channel.

    Args:
      non_media_baseline_values: Optional list of shape
        `(n_non_media_channels,)`. Each element is either a float (which means
        that the fixed value will be used as baseline for the given channel) or
        one of the strings "min" or "max" (which mean that the global minimum or
        maximum value will be used as baseline for the values of the given
        non_media treatment channel). If float values are provided, it is
        expected that they are scaled by population for the channels where
        `model_spec.non_media_population_scaling_id` is `True`. If `None`, the
        `model_spec.non_media_baseline_values` is used, which defaults to the
        minimum value for each non_media treatment channel.

    Returns:
      A tensor of shape `(n_non_media_channels,)` containing the
      baseline values for each non-media treatment channel.
    """
    if non_media_baseline_values is None:
      non_media_baseline_values = (
          self._context.model_spec.non_media_baseline_values
      )

    no_op_scaling_factor = backend.ones_like(self._context.population)[
        :, backend.newaxis, backend.newaxis
    ]
    if self._context.model_spec.non_media_population_scaling_id is not None:
      scaling_factors = backend.where(
          self._context.model_spec.non_media_population_scaling_id,
          self._context.population[:, backend.newaxis, backend.newaxis],
          no_op_scaling_factor,
      )
    else:
      scaling_factors = no_op_scaling_factor

    non_media_treatments_population_scaled = backend.divide_no_nan(
        self._context.non_media_treatments, scaling_factors
    )

    if non_media_baseline_values is None:
      # If non_media_baseline_values is not provided, use the minimum
      # value for each non_media treatment channel as the baseline.
      non_media_baseline_values_filled = [
          constants.NON_MEDIA_BASELINE_MIN
      ] * non_media_treatments_population_scaled.shape[-1]
    else:
      non_media_baseline_values_filled = non_media_baseline_values

    if non_media_treatments_population_scaled.shape[-1] != len(
        non_media_baseline_values_filled
    ):
      raise ValueError(
          "The number of non-media channels"
          f" ({non_media_treatments_population_scaled.shape[-1]}) does not"
          " match the number of baseline values"
          f" ({len(non_media_baseline_values_filled)})."
      )

    baseline_list = []
    for channel in range(non_media_treatments_population_scaled.shape[-1]):
      baseline_value = non_media_baseline_values_filled[channel]

      if baseline_value == constants.NON_MEDIA_BASELINE_MIN:
        baseline_for_channel = backend.reduce_min(
            non_media_treatments_population_scaled[..., channel], axis=[0, 1]
        )
      elif baseline_value == constants.NON_MEDIA_BASELINE_MAX:
        baseline_for_channel = backend.reduce_max(
            non_media_treatments_population_scaled[..., channel], axis=[0, 1]
        )
      elif isinstance(baseline_value, numbers.Number):
        baseline_for_channel = backend.to_tensor(
            baseline_value, dtype=backend.float32
        )
      else:
        raise ValueError(
            f"Invalid non_media_baseline_values value: '{baseline_value}'. Only"
            " float numbers and strings 'min' and 'max' are supported."
        )

      baseline_list.append(baseline_for_channel)

    return backend.stack(baseline_list, axis=-1)

  def linear_predictor_counterfactual_difference_media(
      self,
      *,
      media_transformed: backend.Tensor,
      alpha_m: backend.Tensor,
      ec_m: backend.Tensor,
      slope_m: backend.Tensor,
  ) -> backend.Tensor:
    """Calculates linear predictor counterfactual difference for non-RF media.

    For non-RF media variables (paid or organic), this function calculates the
    linear predictor difference between the treatment variable and its
    counterfactual. "Linear predictor" refers to the output of the hill/adstock
    function, which is multiplied by the geo-level coefficient.

    This function does the calculation efficiently by only calculating calling
    the hill/adstock function if the prior counterfactual is not all zeros.

    Args:
      media_transformed: The output of the hill/adstock function for actual
        historical media data.
      alpha_m: The adstock alpha parameter values.
      ec_m: The adstock ec parameter values.
      slope_m: The adstock hill slope parameter values.

    Returns:
      The linear predictor difference between the treatment variable and its
      counterfactual.
    """
    if self._context.media_tensors.prior_media_scaled_counterfactual is None:
      return media_transformed
    media_transformed_counterfactual = self.adstock_hill_media(
        media=self._context.media_tensors.prior_media_scaled_counterfactual,
        alpha=alpha_m,
        ec=ec_m,
        slope=slope_m,
        decay_functions=self._context.adstock_decay_spec.media,
    )
    # Absolute values is needed because the difference is negative for mROI
    # priors and positive for ROI and contribution priors.
    return backend.absolute(
        media_transformed - media_transformed_counterfactual
    )

  def linear_predictor_counterfactual_difference_rf(
      self,
      *,
      rf_transformed: backend.Tensor,
      alpha_rf: backend.Tensor,
      ec_rf: backend.Tensor,
      slope_rf: backend.Tensor,
  ) -> backend.Tensor:
    """Calculates linear predictor counterfactual difference for RF media.

    For RF media variables (paid or organic), this function calculates the
    linear predictor difference between the treatment variable and its
    counterfactual. "Linear predictor" refers to the output of the hill/adstock
    function, which is multiplied by the geo-level coefficient.

    This function does the calculation efficiently by only calculating calling
    the hill/adstock function if the prior counterfactual is not all zeros.

    Args:
      rf_transformed: The output of the hill/adstock function for actual
        historical media data.
      alpha_rf: The adstock alpha parameter values.
      ec_rf: The adstock ec parameter values.
      slope_rf: The adstock hill slope parameter values.

    Returns:
      The linear predictor difference between the treatment variable and its
      counterfactual.
    """
    if self._context.rf_tensors.prior_reach_scaled_counterfactual is None:
      return rf_transformed
    rf_transformed_counterfactual = self.adstock_hill_rf(
        reach=self._context.rf_tensors.prior_reach_scaled_counterfactual,
        frequency=self._context.rf_tensors.frequency,
        alpha=alpha_rf,
        ec=ec_rf,
        slope=slope_rf,
        decay_functions=self._context.adstock_decay_spec.rf,
    )
    # Absolute values is needed because the difference is negative for mROI
    # priors and positive for ROI and contribution priors.
    return backend.absolute(rf_transformed - rf_transformed_counterfactual)

  def calculate_beta_x(
      self,
      *,
      is_non_media: bool,
      incremental_outcome_x: backend.Tensor,
      linear_predictor_counterfactual_difference: backend.Tensor,
      eta_x: backend.Tensor,
      beta_gx_dev: backend.Tensor,
  ) -> backend.Tensor:
    """Calculates coefficient mean parameter for any treatment variable type.

    The "beta_x" in the function name refers to the coefficient mean parameter
    of any treatment variable. The "x" can represent "m", "rf", "om", or "orf".
    This function can also be used to calculate "gamma_n" for any non-media
    treatments.

    Args:
      is_non_media: Boolean indicating whether the treatment variable is a
        non-media treatment. This argument is used to determine whether the
        coefficient random effects are normal or log-normal. If `True`, then
        random effects are assumed to be normal. Otherwise, the distribution is
        inferred from `self._context.media_effects_dist`.
      incremental_outcome_x: The incremental outcome of the treatment variable,
        which depends on the parameter values of a particular prior or posterior
        draw. The "_x" indicates that this is a tensor with length equal to the
        dimension of the treatment variable.
      linear_predictor_counterfactual_difference: The difference between the
        treatment variable and its counterfactual on the linear predictor scale.
        "Linear predictor" refers to the quantity that is multiplied by the
        geo-level coefficient. For media variables, this is the output of the
        hill/adstock transformation function. For non-media treatments, this is
        simply the treatment variable after centering/scaling transformations.
        This tensor has dimensions for geo, time, and channel.
      eta_x: The random effect standard deviation parameter values. For media
        variables, the "x" represents "m", "rf", "om", or "orf". For non-media
        treatments, this argument should be set to `xi_n`, which is analogous to
        "eta".
      beta_gx_dev: The latent standard normal parameter values of the geo-level
        coefficients. For media variables, the "x" represents "m", "rf", "om",
        or "orf". For non-media treatments, this argument should be set to
        `gamma_gn_dev`, which is analogous to "beta_gx_dev".

    Returns:
      The coefficient mean parameter of the treatment variable, which has
      dimension equal to the number of treatment channels..
    """
    if is_non_media:
      random_effects_normal = True
    else:
      random_effects_normal = (
          self._context.media_effects_dist == constants.MEDIA_EFFECTS_NORMAL
      )
    if self._context.revenue_per_kpi is None:
      revenue_per_kpi = backend.ones(
          [self._context.n_geos, self._context.n_times], dtype=backend.float32
      )
    else:
      revenue_per_kpi = self._context.revenue_per_kpi
    incremental_outcome_gx_over_beta_gx = backend.einsum(
        "...gtx,gt,g,->...gx",
        linear_predictor_counterfactual_difference,
        revenue_per_kpi,
        self._context.population,
        self._context.kpi_transformer.population_scaled_stdev,
    )
    if random_effects_normal:
      numerator_term_x = backend.einsum(
          "...gx,...gx,...x->...x",
          incremental_outcome_gx_over_beta_gx,
          beta_gx_dev,
          eta_x,
      )
      denominator_term_x = backend.einsum(
          "...gx->...x", incremental_outcome_gx_over_beta_gx
      )
      return (incremental_outcome_x - numerator_term_x) / denominator_term_x
    # For log-normal random effects, beta_x and eta_x are not mean & std.
    # The parameterization is beta_gx ~ exp(beta_x + eta_x * N(0, 1)).
    denominator_term_x = backend.einsum(
        "...gx,...gx->...x",
        incremental_outcome_gx_over_beta_gx,
        backend.exp(beta_gx_dev * eta_x[..., backend.newaxis, :]),
    )
    return backend.log(incremental_outcome_x) - backend.log(denominator_term_x)



# Copyright 2025 The Meridian Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the Model Quality Checks."""

import abc
from collections.abc import Sequence
import dataclasses
from typing import Generic, TypeVar
import warnings

from meridian import backend
from meridian import constants
from meridian.analysis import analyzer as analyzer_module
from meridian.analysis.review import configs
from meridian.analysis.review import constants as review_constants
from meridian.analysis.review import results
from meridian.model import model
import numpy as np
import pandas as pd

ConfigType = TypeVar("ConfigType", bound=configs.BaseConfig)
ResultType = TypeVar("ResultType", bound=results.CheckResult)


class BaseCheck(abc.ABC, Generic[ConfigType, ResultType]):
  """A generic, abstract base class for a single, runnable quality check."""

  def __init__(
      self,
      meridian: model.Meridian,
      analyzer: analyzer_module.Analyzer,
      config: ConfigType,
  ):
    self._meridian = meridian
    self._analyzer = analyzer
    self._config = config

  @abc.abstractmethod
  def run(self) -> ResultType:
    """Executes the check.

    The return type uses the generic ResultType, making it specific for each
    subclass.
    """
    raise NotImplementedError()


# ==============================================================================
# Check: Convergence
# ==============================================================================
class ConvergenceCheck(
    BaseCheck[configs.ConvergenceConfig, results.ConvergenceCheckResult]
):
  """Checks for model convergence."""

  def run(self) -> results.ConvergenceCheckResult:
    rhats = self._analyzer.get_rhat()
    with warnings.catch_warnings():
      warnings.filterwarnings("ignore", category=RuntimeWarning)
      max_rhats = {k: np.nanmax(v) for k, v in rhats.items()}

    valid_rhat_items = [
        item for item in max_rhats.items() if not np.isnan(item[1])
    ]
    if not valid_rhat_items:
      return results.ConvergenceCheckResult(
          case=results.ConvergenceCases.CONVERGED,
          details={
              review_constants.RHAT: np.nan,
              review_constants.PARAMETER: np.nan,
              review_constants.CONVERGENCE_THRESHOLD: (
                  self._config.convergence_threshold
              ),
          },
      )

    max_parameter, max_rhat = max(max_rhats.items(), key=lambda item: item[1])

    details = {
        review_constants.RHAT: max_rhat,
        review_constants.PARAMETER: max_parameter,
        review_constants.CONVERGENCE_THRESHOLD: (
            self._config.convergence_threshold
        ),
    }

    # Case 1: Converged.
    if max_rhat < self._config.convergence_threshold:
      return results.ConvergenceCheckResult(
          case=results.ConvergenceCases.CONVERGED,
          details=details,
      )

    # Case 2: Not fully converged, but potentially acceptable.
    elif (
        self._config.convergence_threshold
        <= max_rhat
        < self._config.not_fully_convergence_threshold
    ):
      return results.ConvergenceCheckResult(
          case=results.ConvergenceCases.NOT_FULLY_CONVERGED,
          details=details,
      )

    # Case 3: Not converged and unacceptable.
    else:  # max_rhat >= divergence_threshold
      return results.ConvergenceCheckResult(
          case=results.ConvergenceCases.NOT_CONVERGED,
          details=details,
      )


# ==============================================================================
# Check: Baseline
# ==============================================================================
class BaselineCheck(
    BaseCheck[configs.BaselineConfig, results.BaselineCheckResult]
):
  """Checks for negative baseline probability."""

  def run(self) -> results.BaselineCheckResult:
    prob = self._analyzer.negative_baseline_probability()
    details = {
        review_constants.NEGATIVE_BASELINE_PROB: prob,
        review_constants.NEGATIVE_BASELINE_PROB_FAIL_THRESHOLD: (
            self._config.negative_baseline_prob_fail_threshold
        ),
        review_constants.NEGATIVE_BASELINE_PROB_REVIEW_THRESHOLD: (
            self._config.negative_baseline_prob_review_threshold
        ),
    }
    # Case 1: FAIL
    if prob > self._config.negative_baseline_prob_fail_threshold:
      return results.BaselineCheckResult(
          case=results.BaselineCases.FAIL,
          details=details,
      )
    # Case 2: REVIEW
    elif prob >= self._config.negative_baseline_prob_review_threshold:
      return results.BaselineCheckResult(
          case=results.BaselineCases.REVIEW,
          details=details,
      )
    # Case 3: PASS
    else:
      return results.BaselineCheckResult(
          case=results.BaselineCases.PASS, details=details
      )


# ==============================================================================
# Check: Bayesian Posterior Predictive P-value
# ==============================================================================
class BayesianPPPCheck(
    BaseCheck[configs.BayesianPPPConfig, results.BayesianPPPCheckResult]
):
  """Checks for Bayesian Posterior Predictive P-value."""

  def run(self) -> results.BayesianPPPCheckResult:
    mmm = self._meridian
    analyzer = self._analyzer

    outcome = mmm.kpi
    if mmm.revenue_per_kpi is not None:
      outcome *= mmm.revenue_per_kpi
    total_outcome_actual = np.sum(outcome)

    total_outcome_posterior = analyzer.expected_outcome(
        aggregate_times=True, aggregate_geos=True
    )
    total_outcome_expected = np.asarray(total_outcome_posterior).flatten()

    total_outcome_expected_mean = np.mean(total_outcome_expected)

    bayesian_ppp = np.mean(
        np.abs(total_outcome_expected - total_outcome_expected_mean)
        >= np.abs(total_outcome_actual - total_outcome_expected_mean)
    )

    details = {
        review_constants.BAYESIAN_PPP: bayesian_ppp,
    }

    if bayesian_ppp >= self._config.ppp_threshold:
      return results.BayesianPPPCheckResult(
          case=results.BayesianPPPCases.PASS,
          details=details,
      )
    else:
      return results.BayesianPPPCheckResult(
          case=results.BayesianPPPCases.FAIL,
          details=details,
      )


# ==============================================================================
# Check: Goodness of Fit
# ==============================================================================
def _set_details_from_gof_dataframe(
    details: dict[str, float],
    gof_df: pd.DataFrame,
    geo_granularity: str,
    suffix: str | None = None,
) -> None:
  """Sets the `details` variable of the GoodnessOfFitCheckResult.

  This method takes a DataFrame containing goodness of fit metrics and pivots it
  to a Series, which is then added to the `details` variable of the
  `GoodnessOfFitCheckResult`.

  Args:
    details: A dictionary to store the goodness of fit metrics in.
    gof_df: A DataFrame containing predictive accuracy of the whole data (if
      holdout set is not used) of filtered to a single evaluation set ("all",
      "train", or "test").
    geo_granularity: The geo granularity of the data ("geo" or "national").
    suffix: A suffix to add to the metric names (e.g., "all", "train", "test").
      If None, the metrics are added without a suffix.
  """
  gof_metrics_pivoted = gof_df.pivot(
      index=constants.GEO_GRANULARITY,
      columns=constants.METRIC,
      values=constants.VALUE,
  )
  gof_metrics_series = gof_metrics_pivoted.loc[geo_granularity]
  if suffix is not None:
    details[f"{review_constants.R_SQUARED}_{suffix}"] = gof_metrics_series[
        constants.R_SQUARED
    ]
    details[f"{review_constants.MAPE}_{suffix}"] = gof_metrics_series[
        constants.MAPE
    ]
    details[f"{review_constants.WMAPE}_{suffix}"] = gof_metrics_series[
        constants.WMAPE
    ]
  else:
    details[review_constants.R_SQUARED] = gof_metrics_series[
        constants.R_SQUARED
    ]
    details[review_constants.MAPE] = gof_metrics_series[constants.MAPE]
    details[review_constants.WMAPE] = gof_metrics_series[constants.WMAPE]


class GoodnessOfFitCheck(
    BaseCheck[configs.GoodnessOfFitConfig, results.GoodnessOfFitCheckResult]
):
  """Checks for goodness of fit of the model."""

  def run(self) -> results.GoodnessOfFitCheckResult:
    gof_ds = self._analyzer.predictive_accuracy()
    gof_df = gof_ds.to_dataframe().reset_index()

    geo_granularity = (
        constants.NATIONAL if self._meridian.n_geos == 1 else constants.GEO
    )

    gof_metrics = gof_df[gof_df[constants.GEO_GRANULARITY] == geo_granularity]
    is_holdout = constants.EVALUATION_SET_VAR in gof_df.columns

    details = {}
    case = results.GoodnessOfFitCases.PASS

    if is_holdout:
      for evaluation_set, suffix in [
          (constants.ALL_DATA, review_constants.ALL_SUFFIX),
          (constants.TRAIN, review_constants.TRAIN_SUFFIX),
          (constants.TEST, review_constants.TEST_SUFFIX),
      ]:
        set_metrics = gof_metrics[
            gof_metrics[constants.EVALUATION_SET_VAR] == evaluation_set
        ]
        _set_details_from_gof_dataframe(
            details=details,
            gof_df=set_metrics,
            geo_granularity=geo_granularity,
            suffix=suffix,
        )
        if details[f"{review_constants.R_SQUARED}_{suffix}"] <= 0:
          case = results.GoodnessOfFitCases.REVIEW
    else:
      _set_details_from_gof_dataframe(
          details=details,
          gof_df=gof_metrics,
          geo_granularity=geo_granularity,
          suffix=None,
      )
      if details[review_constants.R_SQUARED] <= 0:
        case = results.GoodnessOfFitCases.REVIEW

    return results.GoodnessOfFitCheckResult(
        case=case,
        details=details,
        is_holdout=is_holdout,
    )


# ==============================================================================
# Check: ROI Consistency
# ==============================================================================
def _format_roi_channels_msg(channels: np.ndarray, direction: str) -> str:
  if channels.size == 0:
    return ""
  plural = "s" if channels.size > 1 else ""
  return (
      f"an unusually {direction} ROI estimate (for channel{plural} "
      f"{', '.join(f'`{channel}`' for channel in channels)})"
  )


def _inf_prior_quantiles_channels(
    channels: np.ndarray,
    lo_roi_quantiles: np.ndarray,
    hi_roi_quantiles: np.ndarray,
) -> np.ndarray:
  """Returns channels with infinite prior quantiles.

  Args:
    channels: The names of the channels.
    lo_roi_quantiles: The lower quantiles of the ROI prior.
    hi_roi_quantiles: The upper quantiles of the ROI prior.

  Returns:
    An array of channel names with infinite prior quantiles.
  """
  inf_mask = np.isinf(lo_roi_quantiles) | np.isinf(hi_roi_quantiles)
  return channels[inf_mask]


@dataclasses.dataclass
class _ROIConsistencyChannelData:
  """A data structure for auxiliary data for the ROI Consistency Check.

  Attributes:
    prior_roi_los: Lower quantile values from ROI priors, corresponding to the
      channels in `all_channels`.
    prior_roi_his: Upper quantile values from ROI priors, corresponding to the
      channels in `all_channels`.
    posterior_means: Mean values of ROI posteriors, corresponding to the
      channels in `all_channels`.
    all_channels: Channel names for which quantile computations were successful;
      channels for which quantiles could not be computed are skipped. They are
      ordered with media channels (`roi_m`) followed by reach and frequency (RF)
      channels (`roi_rf`).
    inf_channels: Channels with infinite prior quantiles.
    low_roi_channels: Channels with posterior means below their prior's lower
      quantile.
    high_roi_channels: Channels with posterior means above their prior's upper
      quantile.
    quantile_not_defined_channels: Channel names for which quantiles could not
      be computed.
    quantile_not_defined_parameters: Parameters for which the quantile method is
      not implemented.
  """

  prior_roi_los: np.ndarray
  prior_roi_his: np.ndarray
  posterior_means: np.ndarray
  all_channels: np.ndarray
  inf_channels: np.ndarray
  low_roi_channels: np.ndarray
  high_roi_channels: np.ndarray
  quantile_not_defined_channels: np.ndarray
  quantile_not_defined_parameters: list[backend.tfd.Distribution] = (
      dataclasses.field(default_factory=list)
  )


def _get_roi_consistency_channel_data(
    prior_rois: Sequence[backend.tfd.Distribution],
    posterior_rois: Sequence[backend.tfd.Distribution],
    channels_names: Sequence[Sequence[str]],
    prior_lower_quantile: float,
    prior_upper_quantile: float,
) -> _ROIConsistencyChannelData:
  """Returns the channel-level data for the ROI Consistency Check.

  Args:
    prior_rois: The ROI priors for all channels, in the same order as
      `channels_names`.
    posterior_rois: The ROI posteriors for all channels, in the same order as
      `channels_names`.
    channels_names: The names of all channels, with media channels (`roi_m`)
      followed by any reach and frequency (RF) channels (`roi_rf`).
    prior_lower_quantile: The lower quantile of the ROI prior.
    prior_upper_quantile: The upper quantile of the ROI prior.

  Returns:
    A _ROIConsistencyChannelData object containing the channel-level data for
    the ROI Consistency Check.
  """

  prior_roi_los_parts = []
  prior_roi_his_parts = []
  posterior_means_parts = []
  all_channels_parts = []
  quantile_not_defined_parameters = []
  quantile_not_defined_channels = []

  for prior_roi, posterior_roi, channels in zip(
      prior_rois, posterior_rois, channels_names
  ):
    try:
      prior_roi_lo = prior_roi.quantile(
          prior_lower_quantile,
      )
      prior_roi_hi = prior_roi.quantile(
          prior_upper_quantile,
      )
      posterior_mean = np.mean(posterior_roi, axis=(0, 1))

      n_channels = len(channels)
      prior_roi_lo = np.broadcast_to(prior_roi_lo, shape=(n_channels,))
      prior_roi_hi = np.broadcast_to(prior_roi_hi, shape=(n_channels,))

      prior_roi_los_parts.append(prior_roi_lo)
      prior_roi_his_parts.append(prior_roi_hi)
      posterior_means_parts.append(posterior_mean)
      all_channels_parts.append(channels)
    except NotImplementedError:
      quantile_not_defined_parameters.append(prior_roi)
      quantile_not_defined_channels.extend(channels)

  if prior_roi_los_parts:
    prior_roi_los = np.concatenate(prior_roi_los_parts)
    prior_roi_his = np.concatenate(prior_roi_his_parts)
    posterior_means = np.concatenate(posterior_means_parts)
    all_channels = np.concatenate(all_channels_parts)
  else:
    prior_roi_los = np.array([])
    prior_roi_his = np.array([])
    posterior_means = np.array([])
    all_channels = np.array([])

  inf_channels = _inf_prior_quantiles_channels(
      channels=all_channels,
      lo_roi_quantiles=prior_roi_los,
      hi_roi_quantiles=prior_roi_his,
  )
  low_roi_channels = all_channels[posterior_means < prior_roi_los]
  high_roi_channels = all_channels[posterior_means > prior_roi_his]

  return _ROIConsistencyChannelData(
      prior_roi_los=prior_roi_los,
      prior_roi_his=prior_roi_his,
      posterior_means=posterior_means,
      all_channels=all_channels,
      inf_channels=inf_channels,
      low_roi_channels=low_roi_channels,
      high_roi_channels=high_roi_channels,
      quantile_not_defined_parameters=quantile_not_defined_parameters,
      quantile_not_defined_channels=np.array(quantile_not_defined_channels),
  )


def _compute_channel_results(
    channel_data: _ROIConsistencyChannelData,
) -> list[results.ROIConsistencyChannelResult]:
  """Returns the channel-level results for the ROI Consistency Check."""

  channel_results = []
  for channel in channel_data.quantile_not_defined_channels:
    case = results.ROIConsistencyChannelCases.QUANTILE_NOT_DEFINED
    channel_results.append(
        results.ROIConsistencyChannelResult(
            case=case,
            details={},
            channel_name=channel,
        )
    )
  for i, channel in enumerate(channel_data.all_channels):
    if channel in channel_data.inf_channels:
      case = results.ROIConsistencyChannelCases.PRIOR_ROI_QUANTILE_INF
    elif channel in channel_data.low_roi_channels:
      case = results.ROIConsistencyChannelCases.ROI_LOW
    elif channel in channel_data.high_roi_channels:
      case = results.ROIConsistencyChannelCases.ROI_HIGH
    else:
      case = results.ROIConsistencyChannelCases.ROI_PASS
    channel_results.append(
        results.ROIConsistencyChannelResult(
            case=case,
            details={
                review_constants.PRIOR_ROI_LO: channel_data.prior_roi_los[i],
                review_constants.PRIOR_ROI_HI: channel_data.prior_roi_his[i],
                review_constants.POSTERIOR_ROI_MEAN: (
                    channel_data.posterior_means[i]
                ),
            },
            channel_name=channel,
        )
    )
  return channel_results


def _compute_aggregate_result(
    channel_data: _ROIConsistencyChannelData,
) -> results.ROIConsistencyCheckResult:
  """Returns the aggregate result for the ROI Consistency Check."""
  channel_results = _compute_channel_results(channel_data=channel_data)

  aggregate_details = {}

  # Channel Case 5: QUANTILE_NOT_DEFINED
  if channel_data.quantile_not_defined_parameters:
    aggregate_details[review_constants.QUANTILE_NOT_DEFINED_MSG] = (
        "The quantile method is not defined for the following parameters:"
        f" {channel_data.quantile_not_defined_parameters}. The ROI"
        " Consistency check cannot be performed for these parameters."
    )
  else:
    aggregate_details[review_constants.QUANTILE_NOT_DEFINED_MSG] = ""

  # Channel Case 4: PRIOR_ROI_QUANTILE_INF
  if channel_data.inf_channels.size > 0:
    aggregate_details[review_constants.INF_CHANNELS_MSG] = (
        "Prior ROI quantiles are infinite for channels:"
        f" {', '.join(channel_data.inf_channels)}"
    )
  else:
    aggregate_details[review_constants.INF_CHANNELS_MSG] = ""

  # Channel Cases 2-3: ROI_LOW, ROI_HIGH
  if (
      channel_data.low_roi_channels.size > 0
      or channel_data.high_roi_channels.size > 0
  ):
    low_msg = _format_roi_channels_msg(channel_data.low_roi_channels, "low")
    high_msg = _format_roi_channels_msg(channel_data.high_roi_channels, "high")

    channels_low_high = " and ".join(filter(None, [low_msg, high_msg]))
    aggregate_details[review_constants.LOW_HIGH_CHANNELS_MSG] = (
        f"We've detected {channels_low_high} where the posterior point"
        " estimate falls into the extreme tail of your custom prior."
    )
  else:
    aggregate_details[review_constants.LOW_HIGH_CHANNELS_MSG] = ""

  if (
      aggregate_details[review_constants.QUANTILE_NOT_DEFINED_MSG]
      or aggregate_details[review_constants.INF_CHANNELS_MSG]
      or aggregate_details[review_constants.LOW_HIGH_CHANNELS_MSG]
  ):
    aggregate_case = results.ROIConsistencyAggregateCases.REVIEW
  else:
    aggregate_case = results.ROIConsistencyAggregateCases.PASS

  return results.ROIConsistencyCheckResult(
      case=aggregate_case,
      details=aggregate_details,
      channel_results=channel_results,
  )


class ROIConsistencyCheck(
    BaseCheck[configs.ROIConsistencyConfig, results.ROIConsistencyCheckResult]
):
  """Checks if ROI posterior mean is in tails of ROI prior."""

  def run(self) -> results.ROIConsistencyCheckResult:
    prior_rois = []
    posterior_rois = []
    channel_names = []
    if (
        constants.MEDIA_CHANNEL
        in self._meridian.inference_data.posterior.coords
    ):
      prior_rois.append(self._meridian.model_spec.prior.roi_m)
      posterior_rois.append(self._meridian.inference_data.posterior.roi_m)
      channel_names.append(
          self._meridian.inference_data.posterior.media_channel.values
      )
    if constants.RF_CHANNEL in self._meridian.inference_data.posterior.coords:
      prior_rois.append(self._meridian.model_spec.prior.roi_rf)
      posterior_rois.append(self._meridian.inference_data.posterior.roi_rf)
      channel_names.append(
          self._meridian.inference_data.posterior.rf_channel.values
      )

    channel_data = _get_roi_consistency_channel_data(
        prior_rois=prior_rois,
        posterior_rois=posterior_rois,
        channels_names=channel_names,
        prior_lower_quantile=self._config.prior_lower_quantile,
        prior_upper_quantile=self._config.prior_upper_quantile,
    )

    return _compute_aggregate_result(channel_data=channel_data)


# ==============================================================================
# Check: Prior-Posterior Shift
# ==============================================================================
def _bootstrap(x: np.ndarray, n_bootstraps: int) -> np.ndarray:
  """Performs non-parametric bootstrap resampling on the columns of x."""
  n_rows, n_cols = x.shape
  x_bs = np.empty((n_bootstraps, n_rows, n_cols))
  for i in range(n_bootstraps):
    col_indices = np.random.choice(n_cols, n_cols, replace=True)
    x_bs[i, :, :] = x[:, col_indices]
  return x_bs


def _calculate_new_statistics_from_samples(
    mmm: model.Meridian, n_bootstraps: int, var_name: str, n_channels: int
) -> dict[str, np.ndarray]:
  """Calculate Mean, Median, Q1, and Q3 from posterior samples."""
  n_chains = len(mmm.inference_data.posterior.coords[constants.CHAIN])
  n_draws = len(mmm.inference_data.posterior.coords[constants.DRAW])
  n_posterior_samples = n_chains * n_draws

  posterior_samples = np.transpose(
      np.reshape(
          mmm.inference_data.posterior.variables[var_name].values,
          (n_posterior_samples, n_channels),
      )
  )
  x = _bootstrap(
      posterior_samples, n_bootstraps
  )  # x is (bootstraps, channels, samples)

  mean = np.mean(x, axis=-1)
  median = np.quantile(x, q=0.5, axis=-1)
  q1 = np.quantile(x, q=0.25, axis=-1)
  q3 = np.quantile(x, q=0.75, axis=-1)

  return {
      review_constants.MEAN: mean,
      review_constants.MEDIAN: median,
      review_constants.Q1: q1,
      review_constants.Q3: q3,
  }


def _get_shifted_mask(
    posterior_stat: np.ndarray, prior_stat: np.ndarray, alpha: float
) -> np.ndarray:
  """Returns a boolean mask indicating which channels have a significant shift."""
  prior_stat_b = prior_stat[np.newaxis, ...]
  shift_1 = np.mean(posterior_stat > prior_stat_b, axis=0) < alpha
  shift_2 = np.mean(posterior_stat < prior_stat_b, axis=0) < alpha
  return shift_1 | shift_2


class PriorPosteriorShiftCheck(
    BaseCheck[
        configs.PriorPosteriorShiftConfig,
        results.PriorPosteriorShiftCheckResult,
    ]
):
  """Checks for a significant shift between prior and posterior of ROI."""

  # Tuple of (channel_results, no_shift_channels)
  _CHANNEL_TYPE_RESULT = tuple[
      list[results.PriorPosteriorShiftChannelResult],
      list[str],
  ]

  def _run_for_channel_type(self, channel_type: str) -> _CHANNEL_TYPE_RESULT:
    """Runs the prior-posterior shift check for a given channel type.

    Args:
      channel_type: The channel type ('media_channel' or 'rf_channel') to run
        the check for.

    Returns:
      A tuple of (`channel_results`, `no_shift_channels`).
    """
    if channel_type not in self._meridian.inference_data.posterior.coords:
      return [], []

    channel_results = []
    no_shift_channels = []

    n_channels = len(
        self._meridian.inference_data.posterior[channel_type].values
    )
    if channel_type == constants.MEDIA_CHANNEL:
      var_name = constants.ROI_M
      prior_dist = self._meridian.model_spec.prior.roi_m
    else:
      var_name = constants.ROI_RF
      prior_dist = self._meridian.model_spec.prior.roi_rf
    prior_stats = {}
    try:
      prior_stats[review_constants.MEAN] = prior_dist.mean()
    except NotImplementedError:
      pass
    try:
      prior_stats[review_constants.MEDIAN] = prior_dist.quantile(0.5)
    except NotImplementedError:
      pass
    try:
      prior_stats[review_constants.Q1] = prior_dist.quantile(0.25)
    except NotImplementedError:
      pass
    try:
      prior_stats[review_constants.Q3] = prior_dist.quantile(0.75)
    except NotImplementedError:
      pass

    post_stats = _calculate_new_statistics_from_samples(
        self._meridian, self._config.n_bootstraps, var_name, n_channels
    )

    alpha = self._config.alpha
    any_shift = np.zeros(n_channels, dtype=bool)
    for key in prior_stats:
      prior_stat = prior_stats[key]
      post_stat = post_stats[key]
      current_shift = _get_shifted_mask(post_stat, prior_stat, alpha)
      any_shift = any_shift | current_shift

    channel_names = self._meridian.inference_data.posterior[channel_type].values
    for i, channel_name in enumerate(channel_names):
      shifted = any_shift[i]
      case = (
          results.PriorPosteriorShiftChannelCases.SHIFT
          if shifted
          else results.PriorPosteriorShiftChannelCases.NO_SHIFT
      )
      if not shifted:
        no_shift_channels.append(channel_name)
      channel_results.append(
          results.PriorPosteriorShiftChannelResult(
              case=case, details={}, channel_name=channel_name
          )
      )
    return channel_results, no_shift_channels

  def _aggregate(
      self,
      *channel_type_results: _CHANNEL_TYPE_RESULT,
  ) -> results.PriorPosteriorShiftCheckResult:
    """Aggregates results from multiple channel types."""
    channel_results = []
    no_shift_channels = []
    for results_part, channels_part in channel_type_results:
      channel_results.extend(results_part)
      no_shift_channels.extend(channels_part)

    if no_shift_channels:
      agg_case = results.PriorPosteriorShiftAggregateCases.REVIEW
      final_details = {
          "channels_str": ", ".join(
              f"`{channel}`" for channel in no_shift_channels
          )
      }
    else:
      agg_case = results.PriorPosteriorShiftAggregateCases.PASS
      final_details = {}

    return results.PriorPosteriorShiftCheckResult(
        case=agg_case, details=final_details, channel_results=channel_results
    )

  def run(self) -> results.PriorPosteriorShiftCheckResult:
    np.random.seed(self._config.seed)
    media_results = self._run_for_channel_type(constants.MEDIA_CHANNEL)
    rf_results = self._run_for_channel_type(constants.RF_CHANNEL)
    return self._aggregate(media_results, rf_results)
